
    #	Sanity check CSV transit canon
    
    #	Input:	transits.csv  all_fine.txt.gz
    #	Output: Standard output
    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/

    use strict;
    use warnings;
    
    #	Performing the Pgase 2 test exhaustively in Perl without
    #	additional cleverness if prohibitive--it would take
    #	almost a month on a 3.4 GHz Pentium 4 machine.  So,
    #	for the moment we perform statistical quality control,
    #	pseudorandomly selecting a subset of the records to
    #	be tested.  If there's a problem, this is likely to
    #	encounter it at least once.  Set $testEvery to the
    #	reciprocal of the fraction of records you wish to test.
    
    my $testEvery = 157;    	    # 157 runs about 3 hours on a 3.4 GHz Pentium 4

    my $l;
    my $lineno = 0;
    my $errors = 0;
    my $laststart = -1E100; 	    # Last transit end time
    
    #	Planet orbital periods in days.  This is used to test
    #	for consecutive transits of the same planet which are
    #	implausibly close together.  There's no to obsess about
    #	the precision of these values; it's fine if they're a
    #	little on the low side.
    my %orbitalPeriod = (
	Mercury => 87.94,
	Venus => 224.7,
	Earth => 365.2,
	Moon => 365.2,
	Mars => 687.0,
	Jupiter => 4331,
	Saturn => 10747,
	Uranus => 30589,
	Neptune => 59800,
	Pluto => 90588
    );
    
    #	Time last transit started.  Indexed by a key consisting
    #	of "FromPlanet-TransitingPlanet".
    my %laststart;
    
    my %records;
    
    my ($Pof, $Pfrom, $Pjd);
    
    open(FI, "<transits.csv");
    while ($l = <FI>) {
    	$lineno++;
    	$l =~ s/\s+//;
	$l =~ m/^([^\,]+),([^\,]+),([^\,]+),([^\,]+),([^\,]+),([^\,]+),([^\,]+)$/;
    	my ($transitof, $from, $start, $max, $end, $sunsubtends, $transangle) =
	    ($1, $2, $3, $4, $5, $6, $7);
	
	#   Check for legacy "optimisation" of numbers which turned
	#   "0.0" into ".".
	if ($transangle eq '.') {
	    print("$lineno: Unpatched grazing transit angle ($l)\n");
	    $errors++;
    	}
	
	#   Angular distance of transiting planet must be <= angle
	#   subtended by the Sun.
	if ($transangle > $sunsubtends) {
	    print("$lineno: Transangle greater than Sunsubtends ($l)\n");
	    $errors++;
	}
	
	if ($end < $laststart) {
	    print("$lineno: Start before after last start time $laststart ($l)\n");
	    $errors++;
	}
	$laststart = $start;
	
	#   Start time must be <= end time
	if ($start > $end) {
 	    print("$lineno: Start time after end time ($l)\n");
	    $errors++;
   	}
	
	#   Maximum transit time must be >= start time...
	if ($max < $start) {
 	    print("$lineno: Max time before start time ($l)\n");
	    $errors++;
   	}
	
	#   ...and <= end time
	if ($max > $end) {
 	    print("$lineno: Max time after end time ($l)\n");
	    $errors++;
   	}
	
	#   No legitimate transit has a duration of two days or more
	if (($end - $start) > 2) {
	    print("$lineno: Transit duration (" .
	    	sprintf("%.2f", $end - $start) . " days) too long ($l)\n");
	    $errors++;
	}
	
	#   Consecutive transits of the same planet viewed from a given
	#   planet cannot occur in less than the orbital period of the
	#   inner planet.
	my $tid = "$transitof-$from";
	if (defined($laststart{$tid})) {
	    if (($start - $laststart{$tid}) < $orbitalPeriod{$transitof}) {
	    	print("$lineno: Period (" .
		    sprintf("%.3f", $start - $laststart{$tid}) .
		    " days) between transits of $transitof less than its orbital period ($l)\n");
		$errors++;
	    }
	}
	$laststart{$tid} = $start;
	
	#   Save start and end of event for completeness/uniqueness test in Phase 2
	if (!defined($records{$tid})) {
	    $records{$tid} = [ ];
	}
	push(@{$records{$tid}}, [ $start, $end ]);
    }
    close(FI);
    
    #	Phase 2:  "No transit left behind".  Verify that each transit event
    #	    	  identified in the transits.c output is contained within
    #	    	  one and only one CSV item in the canon.
    
    if ($errors == 0) {
	open(FI, "zcat ../Integration_Results/all_fine.txt.gz |") ||
	    die("Cannot open pipe from ../Integration_Results/all_fine.txt.gz");

    	my $lpd = -1E100;   	    # Last progress century printed
	my ($tested, $skipped) = (0, 0);
	my $ltime = time();
	while ($l = <FI>) {

	    if (!parse_transit($l)) {
    		next;
	    }
	    	    
    	    if (int(rand($testEvery)) == int($testEvery / 2)) {
    		$tested++;
		my $tl = \@{$records{"$Pof-$Pfrom"}};
		my $n = $#{$records{"$Pof-$Pfrom"}};
    		my $contained = 0;
		for (my $i = 0; $i <= $n; $i++) {
    	    	    if (($Pjd >= $$tl[$i][0]) && ($Pjd <= $$tl[$i][1])) {
			$contained++;
		    }
		}
		if (($Pjd - $lpd) > 365242.5) {
	    	    my $ntime = time();
		    my $dtime = int($ntime - $ltime);
	    	    print("...($contained) $tested/$skipped [$dtime sec] $l");
		    $lpd = $Pjd;
		    $ltime = $ntime;
    		}
		if ($contained != 1) {
	    	    if ($contained == 0) {
			print("Record missing from transit database:\n    $l");
		    } else {
			print("Item present in $contained items in transit database:\n    $l");
		    }
		    $errors++;
		}
    	    } else {
	    	$skipped++;
	    }
	}
	close(FI);
print("Tested $tested records, skipped $skipped.\n");
    } else {
    	print("Errors in phase 1.  Skipping phase 2.\n");
    }

    if ($errors > 0) {
    	print("$errors error(s) found.\n");
    }

    #	Parse a transit event record into global field variables
    
    sub parse_transit {
	my ($l) = @_;
	
    	$l =~ s/\s+$//;
	if ($l =~ m/^Transit of (\w+)\s+from (\w+)\s+(\S+)\s+(.*)$/) {
	    ($Pof, $Pfrom, $Pjd) = ($1, $2, $3);
	    return 1;
	}
	return 0;
    } 
