/*

    	      Julian Day Utilities
	    
	         by John Walker
	    http://www.fourmilab.ch/
	    
       This program is in the public domain.
    
    For details, please see: http://www.fourmilab.ch/documents/canon_transits/

*/

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include "mathutil.h"
#include "julian.h"

static const double GREGORIAN_EPOCH = 1721425.5;

/*  LEAP_GREGORIAN  --  Is a given year in the Gregorian calendar
    	    	        a leap year ?
*/

int leap_gregorian(const int year)
{
    return ((year % 4) == 0) &&
            (!(((year % 100) == 0) && ((year % 400) != 0)));
}

/*  GREGORIAN_TO_JD  --  Determine Julian day number from
    	    	    	 Gregorian calendar date.
*/

double gregorian_to_jd(const int year, const int month, const int day)
{
    
    return (GREGORIAN_EPOCH - 1) +
           (365 * (year - 1)) +
           floor((year - 1) / 4.0) +
           (-floor((year - 1) / 100.0)) +
           floor((year - 1) / 400.0) +
           ((((367 * month) - 362) / 12) +
           ((month <= 2) ? 0 :
                               (leap_gregorian(year) ? -1 : -2)
           ) +
           day);
}

/*  JD_TO_GREGORIAN  --  Calculate Gregorian calendar date from Julian day
*/

void jd_to_gregorian(const double jd, int *yy, int *mm, int *dd) {
    double wjd, depoch, quadricent, dqc, cent, dcent, quad, dquad,
        yindex, yearday, leapadj;
    int year, month, day;

    wjd = floor(jd - 0.5) + 0.5;
    depoch = wjd - GREGORIAN_EPOCH;
    quadricent = floor(depoch / 146097);
    dqc = mod(depoch, 146097);
    cent = floor(dqc / 36524);
    dcent = mod(dqc, 36524);
    quad = floor(dcent / 1461);
    dquad = mod(dcent, 1461);
    yindex = floor(dquad / 365);
    year = (int) ((quadricent * 400) + (cent * 100) + (quad * 4) + yindex);
    if (!((cent == 4) || (yindex == 4))) {
        year++;
    }
    yearday = wjd - gregorian_to_jd(year, 1, 1);
    leapadj = ((wjd < gregorian_to_jd(year, 3, 1)) ? 0
                                                  :
                  (leap_gregorian(year) ? 1 : 2)
              );
    month = (int) ((((yearday + leapadj) * 12) + 373) / 367);
    day = (int) (wjd - gregorian_to_jd(year, month, 1)) + 1;

    *yy = year;
    *mm = month;
    *dd = day;
}

/*  LEAP_JULIAN  --  Determine whether a Julian year number is a leap year
*/

static const double JULIAN_EPOCH = 1721423.5;

int leap_julian(const int year)
{
    return (year % 4) == ((year > 0) ? 0 : 3);
}

/*  JULIAN_TO_JD  --  Determine Julian day number from Julian calendar date
*/

double julian_to_jd(int year, int month, const int day)
{

    /* Adjust negative common era years to the zero-based notation we use.  */

    if (year < 1) {
        year++;
    }

    /* Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page 61 */

    if (month <= 2) {
        year--;
        month += 12;
    }

    return ((floor((365.25 * (year + 4716))) +
            floor((30.6001 * (month + 1))) +
            day) - 1524.5);
}

/*  JD_TO_JULIAN  --  Calculate Julian calendar date from Julian day
*/

void jd_to_julian(double jd, int *yy, int *mm, int *dd) {
    double z, a, b, c, d, e, year, month, day;

    jd += 0.5;
    z = floor(jd);

    a = z;
    b = a + 1524;
    c = floor((b - 122.1) / 365.25);
    d = floor(365.25 * c);
    e = floor((b - d) / 30.6001);

    month = (int) ((e < 14) ? (e - 1) : (e - 13));
    year = (int) ((month > 2) ? (c - 4716) : (c - 4715));
    day = b - d - ((int) (30.6001 * e));

    /*  If year is less than 1, subtract one to convert from
        a zero based date system to the common era system in
        which the year -1 (1 B.C.E) is followed by year 1 (1 C.E.).  */

    if (year < 1) {
        year--;
    }

    *yy = year;
    *mm = month;
    *dd = day;
}

/*  UNIX_TIME_TO_JD  --  Compute Julian day from Unix time() value.
    	    	    	 Note that we specify the time() value as a
			 double, which permits values outside the range
			 which fit in an int as well as fractional seconds.  */

static const double UNIX_TIME_EPOCH = 2440587.5;    /* Unix time() value epoch */

double unix_time_to_jd(const double utime)
{
    return UNIX_TIME_EPOCH + (utime / (60 * 60 * 24));
}

/*  JD_TO_UNIX_TIME  --  Compute Unix time() value from Julian day.
    	    	    	 The value is returned as a double but is always
			 rounded to the nearest second.  */

double jd_to_unix_time(const double jd)
{
    return floor(((jd - UNIX_TIME_EPOCH) * (60 * 60 * 24)) + 0.5);
}

/*  JD_TO_CIVIL_TIME  --  Convert Julian time to hour, minutes, and seconds.  */

void jd_to_civil_time(double j, int *h, int *m, int *s)
{
    int ij;

    j += 0.5;		      /* Astronomical to civil */
    ij = (int) ((j - floor(j)) * 86400.0);
    *h = (int) (ij / 3600L);
    *m = (int) ((ij / 60L) % 60L);
    *s = (int) (ij % 60L);
}

/*  JD_TO_WEEKDAY  --  Determine day of week from Julian date.
    	    	       0 = Sunday, 1 = Monday, ... 6 = Saturday.  */
	       
const char *weekday_names[7] = { "Sunday", "Monday", "Tuesday", "Wednesday",
                     	         "Thursday", "Friday", "Saturday" };
	       
int jd_to_weekday(const double j)
{
    int ij = (int) (j + 1.5);
    
    ij %= 7;
    
    return (ij < 0) ? (7 + ij) : ij;
}
